import autograd.numpy as np

from .expression import Expr
from .constraint import Constraint
from .set import Set


class Variable:
    __array_priority__ = 100

    _ids = 0

    def __init__(
        self,
        shape=(1,),
        name=None,
        symmetric=False,
        PSD=False,
        NSD=False,
        pos=False,
        neg=False,
        integer=False,
    ):
        assert isinstance(shape, tuple), "Shape must be a tuple"
        assert len(shape) > 0, "Shape must be non-empty"
        assert all(isinstance(s, int) for s in shape), (
            "Shape must be a tuple of integers"
        )
        assert all(s > 0 for s in shape), "Shape must be a tuple of positive integers"
        assert len(shape) == 1 or len(shape) == 2, (
            "Shape must be a tuple of length 1 or 2"
        )

        self.name = name if name else f"x{Variable._ids}"
        self.shape = shape
        self.size = int(np.prod(shape)) if shape else 1
        self._value = None
        self._id = Variable._ids
        Variable._ids += 1

        self.constraints = []

        if symmetric:
            assert len(shape) == 2, "Symmetric variable must be a square matrix"
            assert shape[0] == shape[1], "Symmetric variable must be a square matrix"
            U_inds = np.triu_indices(shape[0], k=1)
            L_inds = np.tril_indices(shape[0], k=-1)
            self.constraints.append(Constraint(self[U_inds], "==", self[L_inds]))

        if PSD:
            assert len(shape) == 2, "PSD variable must be a square matrix"
            assert shape[0] == shape[1], "PSD variable must be a square matrix"
            assert not neg, "PSD variable cannot be negative"
            self.constraints.append(Constraint(self, ">>", 0))

        if NSD:
            assert len(shape) == 2, "NSD variable must be a square matrix"
            assert shape[0] == shape[1], "NSD variable must be a square matrix"
            assert not pos, "NSD variable cannot be positive"
            self.constraints.append(Constraint(self, "<<", 0))

        if pos:
            assert not neg and not NSD, "Positive variable cannot be NSD or negative"
            self.constraints.append(Constraint(self, ">=", 0))

        if neg:
            assert not PSD and not pos, "Negative variable cannot be PSD or positive"
            self.constraints.append(Constraint(self, "<=", 0))

        if integer:
            raise NotImplementedError("Integer variables are not supported yet")

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = np.array(val).reshape(self.shape)

    def __repr__(self):
        return f"Var({self.name}, shape={self.shape})"

    @property
    def T(self):
        return Expr("transpose", self)

    def __add__(self, other):
        return Expr("add", self, other)

    def __radd__(self, other):
        return Expr("add", other, self)

    def __sub__(self, other):
        return Expr("sub", self, other)

    def __rsub__(self, other):
        return Expr("sub", other, self)

    def __mul__(self, other):
        return Expr("mul", self, other)

    def __rmul__(self, other):
        return Expr("mul", other, self)

    def __matmul__(self, other):
        return Expr("matmul", self, other)

    def __rmatmul__(self, other):
        return Expr("matmul", other, self)

    def __truediv__(self, other):
        return Expr("div", self, other)

    def __pow__(self, other):
        return Expr("pow", self, other)

    def __neg__(self):
        return Expr("neg", self)

    def __getitem__(self, key):
        return Expr("getitem", self, key)

    def __ge__(self, other):
        return Constraint(self, ">=", other)

    def __le__(self, other):
        return Constraint(self, "<=", other)

    def __eq__(self, other):
        return Constraint(self, "==", other)

    def __rshift__(self, other):
        return Constraint(self, ">>", other)

    def __lshift__(self, other):
        return Constraint(self, "<<", other)

    def __xor__(self, other):
        assert isinstance(other, Set), "Set must be a Set object"
        return other.constrain(self)



from typing import Callable

from .expression import Expr
from .variable import Variable


def collect_vars(expr, vars):
    if isinstance(expr, Variable):
        vars.append(expr)
    elif isinstance(expr, Expr):
        collect_vars(expr.left, vars)
        if expr.right is not None:
            collect_vars(expr.right, vars)


def eval_expression(expr, var_dict, use_value=False):
    if isinstance(expr, Variable):
        if use_value:
            return expr.value
        else:
            return var_dict[expr.name]

    elif isinstance(expr, Expr):
        left_eval = eval_expression(expr.left, var_dict, use_value)
        right_eval = (
            eval_expression(expr.right, var_dict, use_value)
            if expr.right is not None
            else None
        )

        if expr.op == "add":
            return left_eval + right_eval
        elif expr.op == "sub":
            return left_eval - right_eval
        elif expr.op == "mul":
            return left_eval * right_eval
        elif expr.op == "div":
            return left_eval / right_eval
        elif expr.op == "pow":
            return left_eval**right_eval
        elif expr.op == "neg":
            return -left_eval
        elif expr.op == "matmul":
            return left_eval @ right_eval
        elif expr.op == "getitem":
            return left_eval[right_eval]
        elif expr.op == "transpose":
            return left_eval.T
        elif isinstance(expr, Callable):
            return expr(left_eval) if right_eval is None else expr(left_eval, right_eval)
        else:
            raise NotImplementedError(expr.op)

    else:
        return expr



import autograd.numpy as np

from ..expression import Expr


class maximum(Expr):
    def __init__(self, left, right):
        super().__init__("maximum", left, right)

    def __call__(self, x, y):
        return np.maximum(x, y)



import autograd.numpy as np

from ..expression import Expr
from ..overrides import svd


class PolarDecomposition(Expr):
    def __init__(self, left):
        super().__init__("polar_decomp", left)

    def __call__(self, x):
        U, _, Vt = svd(x, full_matrices=False)
        det_UVt = np.linalg.det(U @ Vt)
        S = np.ones((x.shape[0],))
        S[-1] = np.sign(det_UVt)
        return U @ np.diag(S) @ Vt



__all__ = [
    "det",
    "norm",
    "PolarDecomposition",
    "sum",
    "trace",
    "maximum",
]

from .det import det
from .norm import norm
from .polar import PolarDecomposition
from .sum import sum
from .trace import trace
from .maximum import maximum



__all__ = [
    "Variable",
    "Constraint",
    "Set",
    "Expr",
    "Problem",
    "Maximize",
    "Minimize",
    "SLSQP",
    "IPOPT",
    "COBYLA",
    "NELDER_MEAD",
    "BFGS",
    "LBFGSB",
    "TNC",
    "TRUST_CONSTR",
    "det",
    "norm",
    "sum",
    "trace",
    "maximum",
    "Function",
]

from .variable import Variable
from .constraint import Constraint
from .set import Set
from .expression import Expr
from .problem import Problem, Maximize, Minimize
from .constants import (
    SLSQP,
    IPOPT,
    COBYLA,
    NELDER_MEAD,
    BFGS,
    LBFGSB,
    TNC,
    TRUST_CONSTR,
)

from .atoms import det, norm, sum, trace, maximum
from .constructs import Function



from typing import Optional
from dataclasses import dataclass
import time

import autograd.numpy as np
from autograd import grad, jacobian
from scipy.optimize import minimize

from .constants import SLSQP
from .parser import collect_vars, eval_expression


@dataclass(frozen=True)
class SolverStats:
    solver_name: str
    solve_time: Optional[float] = None
    setup_time: Optional[float] = None
    num_iters: Optional[int] = None


class Minimize:
    def __init__(self, expr):
        self.expr = expr


class Maximize:
    def __init__(self, expr):
        self.expr = -expr


class Problem:
    def __init__(self, objective, constraints=[]):
        self.objective = objective
        self.constraints = constraints

        self.vars = []
        self.var_shapes = {}
        self.var_slices = {}
        self.total_size = 0

        all_vars = []
        collect_vars(objective.expr, all_vars)
        for c in constraints:
            collect_vars(c.left, all_vars)
            collect_vars(c.right, all_vars)

        variable_constraints = []
        for v in all_vars:
            variable_constraints.extend(v.constraints)
        self.constraints += variable_constraints

        self.vars = []
        for v in all_vars:
            if v.name not in self.var_shapes:
                self.var_shapes[v.name] = v.shape
                self.var_slices[v.name] = (self.total_size, self.total_size + v.size)
                self.total_size += v.size
                self.vars.append(v)

        self.status = None
        self.solver_stats = None

    def solve(self, solver=SLSQP, solver_options=None, pre_solve=False):
        start_setup_time = time.time()

        x0 = np.ones(self.total_size)
        var_names = [v.name for v in self.vars]
        var_shapes = self.var_shapes
        var_slices = self.var_slices

        for v in self.vars:
            v_start, v_end = self.var_slices[v.name]
            if v.value is None:
                x0[v_start:v_end] = np.zeros(v.size)
            else:
                x0[v_start:v_end] = np.ravel(v.value)

        def unpack(x):
            var_dict = {}
            for v in var_names:
                start, end = var_slices[v]
                shape = var_shapes[v]
                val = x[start:end]
                var_dict[v] = val if not shape else val.reshape(shape)
            return var_dict

        def obj_func(x):
            var_dict = unpack(x)
            return eval_expression(self.objective.expr, var_dict)

        jac_func = grad(obj_func)

        cons = []
        for c in self.constraints:

            def make_con_fun(c):
                def con_fun(x):
                    var_dict = unpack(x)
                    lval = eval_expression(c.left, var_dict)
                    rval = eval_expression(c.right, var_dict)
                    res = lval - rval if c.op in [">=", "==", ">>"] else rval - lval
                    if c.op in [">>", "<<"]:
                        return np.real(np.ravel(np.linalg.eig(res)[0]))
                    return np.ravel(res)

                return con_fun, jacobian(con_fun)

            ctype = "eq" if c.op == "==" else "ineq"
            con_fun, con_jac = make_con_fun(c)
            cons.append({"type": ctype, "fun": con_fun, "jac": con_jac})

        setup_time = time.time() - start_setup_time

        solve_time = 0.0
        if pre_solve:

            def dummy_func(_):
                return 0.0

            def dummy_jac(x):
                return np.zeros_like(x)

            start_time = time.time()
            results_constraints = minimize(
                dummy_func,
                x0,
                jac=dummy_jac,
                constraints=cons,
                method=solver,
                options=solver_options,
            )
            x0 = results_constraints.x
            solve_time += time.time() - start_time

        start_time = time.time()
        result = minimize(
            obj_func,
            x0,
            jac=jac_func,
            constraints=cons,
            method=solver,
            options=solver_options,
        )
        solve_time += time.time() - start_time

        self.status = result.status
        self.solver_stats = SolverStats(
            solver_name=solver,
            solve_time=solve_time,
            setup_time=setup_time,
            num_iters=result.nit,
        )

        sol_vars = unpack(result.x)

        for v in self.vars:
            v.value = sol_vars[v.name]



import autograd.numpy as np

from ..expression import Expr


class sum(Expr):
    def __init__(self, left, axis=None):
        self.axis = axis
        super().__init__("sum", left)

    def __call__(self, x, axis=None):
        return np.sum(x, axis=axis)



from ..constraint import Constraint
from ..set import Set
from ..atoms.polar import PolarDecomposition


class SO(Set):
    def __init__(self, n):
        super().__init__(f"SO({n})")
        self.n = n

    def constrain(self, var):
        assert var.shape == (self.n, self.n)
        return Constraint(var, "==", PolarDecomposition(var))



from .constraint import Constraint
from .set import Set


class Expr:
    __array_priority__ = 100

    def __init__(self, op, left, right=None):
        self.op = op
        self.left = left
        self.right = right

    @property
    def value(self):
        from .parser import eval_expression

        return eval_expression(self, None, use_value=True)

    @property
    def T(self):
        return Expr("transpose", self)

    def __add__(self, other):
        return Expr("add", self, other)

    def __radd__(self, other):
        return Expr("add", other, self)

    def __sub__(self, other):
        return Expr("sub", self, other)

    def __rsub__(self, other):
        return Expr("sub", other, self)

    def __mul__(self, other):
        return Expr("mul", self, other)

    def __rmul__(self, other):
        return Expr("mul", other, self)

    def __matmul__(self, other):
        return Expr("matmul", self, other)

    def __rmatmul__(self, other):
        return Expr("matmul", other, self)

    def __truediv__(self, other):
        return Expr("div", self, other)

    def __pow__(self, other):
        return Expr("pow", self, other)

    def __neg__(self):
        return Expr("neg", self)

    def __getitem__(self, key):
        return Expr("getitem", self, key)

    def __ge__(self, other):
        return Constraint(self, ">=", other)

    def __le__(self, other):
        return Constraint(self, "<=", other)

    def __eq__(self, other):
        return Constraint(self, "==", other)

    def __rshift__(self, other):
        return Constraint(self, ">>", other)

    def __lshift__(self, other):
        return Constraint(self, "<<", other)

    def __xor__(self, other):
        assert isinstance(other, Set), "Set must be a Set object"
        return other.constrain(self)



import autograd.numpy as np

from ..expression import Expr


class norm(Expr):
    def __init__(self, left, ord=2, axis=None):
        self.ord = ord
        self.axis = axis
        super().__init__("norm", left)

    def __call__(self, x):
        return np.linalg.norm(x, ord=self.ord, axis=self.axis)



import autograd.numpy as np

from ..expression import Expr


class det(Expr):
    def __init__(self, left):
        super().__init__("det", left)

    def __call__(self, x):
        return np.linalg.det(x)



import autograd.numpy as np

from ..expression import Expr


class trace(Expr):
    def __init__(self, left, offset=0):
        self.offset = offset
        super().__init__("trace", left)

    def __call__(self, x):
        return np.trace(x, offset=self.offset)



from typing import Callable

from autograd import jacobian
from autograd.extend import defvjp
from scipy.optimize import approx_fprime

from ..expression import Expr


class Function(Expr):
    def __init__(self, func, left, jac="numerical"):
        self.func = func

        if jac == "numerical":
            self.jac = self._numerical_diff
        elif jac == "autograd":
            self.jac = jacobian(self.__call__)
        elif isinstance(jac, Callable):
            self.jac = jac
        else:
            raise ValueError(f"Invalid jacobian: {jac}")

        defvjp(self.__call__, self.jac)

        super().__init__("func", left)

    def __call__(self, x):
        return self.func(x)

    def _numerical_diff(self, ans, x):
        return lambda g: approx_fprime(x, self.__call__, epsilon=1e-8) * g
